@@ -43,6 +43,11 @@ except NameError:
     # Python 3 support
     str_class = str
 
+# backward compatibility for basestring for python < 2.3
+try:
+    basestring
+except NameError:
+    basestring = str
 
 # Flag for testing
 debug = False
@@ -66,6 +71,7 @@ templatecommand ::= <blockrule>
     | <escapedexpression>
     | <expression>
     | <partial>
+    | <rawblock>
 start ::= '{' '{'
 finish ::= '}' '}'
 comment ::= <start> '!' (~(<finish>) <anything>)* <finish> => ('comment', )
@@ -84,10 +90,11 @@ alt_inner ::= <spaces> ('^' | 'e' 'l' 's' 'e') <spaces> <finish>
 partial ::= <start> '>' <partial_inner>:i => ('partial',) + i
 path ::= ~('/') <pathseg>+:segments => ('path', segments)
 kwliteral ::= <safesymbol>:s '=' (<literal>|<path>|<subexpression>):v => ('kwparam', s, v)
-literal ::= (<string>|<integer>|<boolean>|<null>|<undefined>):thing => ('literalparam', thing)
+literal ::= (<string>|<float>|<integer>|<boolean>|<null>|<undefined>):thing => ('literalparam', thing)
 string ::= '"' <notdquote>*:ls '"' => u'"' + u''.join(ls) + u'"'
     | "'" <notsquote>*:ls "'" => u"'" + u''.join(ls) + u"'"
 integer ::= '-'?:sign <digit>+:ds => int((sign if sign else '') + ''.join(ds))
+float ::= '-'?:sign <digit>+:ds '.':ss <digit>+:fs => float((sign if sign else '') + ''.join(ds) + ss + ''.join(fs))
 boolean ::= <false>|<true>
 false ::= 'f' 'a' 'l' 's' 'e' => False
 true ::= 't' 'r' 'u' 'e' => True
@@ -121,8 +128,12 @@ symbolfinish :expected ::= <start> '/' <symbol>:found ?(found == expected) <fini
 blockrule ::= <start> '#' <block_inner>:i
       <template>:t <alttemplate>:alt_t <symbolfinish i[0]> => ('block',) + i + (t, alt_t)
     | <start> '^' <block_inner>:i
-      <template>:t <alttemplate>:alt_t <symbolfinish i[0]> => ('invertedblock',) + i + (t,alt_t)
+      <template>:t <alttemplate>:alt_t <symbolfinish i[0]> => ('invertedblock',) + i + (t, alt_t)
 alttemplate ::= (<start> <alt_inner> <template>)?:alt_t => alt_t or []
+rawblockstart ::= <start> <start> <block_inner>:i <finish> => i
+rawblockfinish :expected ::= <start> <symbolfinish expected> <finish>
+rawblock ::= <rawblockstart>:i (~(<rawblockfinish i[0]>) <anything>)*:r <rawblockfinish i[0]>
+    => ('rawblock',) + i + (''.join(r),)
 """
 
 # this grammar compiles the template to python
@@ -135,6 +146,7 @@ rule ::= <literal>
     | <comment>
     | <block>
     | <invertedblock>
+    | <rawblock>
     | <partial>
 block ::= [ "block" <anything>:symbol [<arg>*:arguments] [<compile>:t] [<compile>?:alt_t] ] => builder.add_block(symbol, arguments, t, alt_t)
 comment ::= [ "comment" ]
@@ -142,11 +154,12 @@ literal ::= [ ( "literal" | "newline" | "whitespace" ) :value ] => builder.add_l
 expand ::= [ "expand" <path>:value [<arg>*:arguments]] => builder.add_expand(value, arguments)
 escapedexpand ::= [ "escapedexpand" <path>:value [<arg>*:arguments]] => builder.add_escaped_expand(value, arguments)
 invertedblock ::= [ "invertedblock" <anything>:symbol [<arg>*:arguments] [<compile>:t] [<compile>?:alt_t] ] => builder.add_invertedblock(symbol, arguments, t, alt_t)
+rawblock ::= [ "rawblock" <anything>:symbol [<arg>*:arguments] <anything>:raw ] => builder.add_rawblock(symbol, arguments, raw)
 partial ::= ["partial" <complexarg>:symbol [<arg>*:arguments]] => builder.add_partial(symbol, arguments)
 path ::= [ "path" [<pathseg>:segment]] => ("simple", segment)
- | [ "path" [<pathseg>+:segments] ] => ("complex", u"resolve(context, '"  + u"', '".join(segments) + u"')" )
-complexarg ::= [ "path" [<pathseg>+:segments] ] => u"resolve(context, '"  + u"', '".join(segments) + u"')"
-    | [ "subexpr" ["path" <pathseg>:name] [<arg>*:arguments] ] => u'resolve_subexpr(helpers, "' + name + '", context' + (u', ' + u', '.join(arguments) if arguments else u'') + u')'
+ | [ "path" [<pathseg>+:segments] ] => ("complex", u"resolve(context, u'" + u"', u'".join(segments) + u"')" )
+complexarg ::= [ "path" [<pathseg>+:segments] ] => u"resolve(context, u'" + u"', u'".join(segments) + u"')"
+    | [ "subexpr" ["path" <anything>:name] [<arg>*:arguments] ] => u'resolve_subexpr(helpers, "' + '.'.join(name) + '", context' + (u', ' + u', '.join(arguments) if arguments else u'') + u')'
     | [ "literalparam" <anything>:value ] => {str_class}(value)
 arg ::= [ "kwparam" <anything>:symbol <complexarg>:a ] => {str_class}(symbol) + '=' + a
     | <complexarg>
@@ -216,8 +229,15 @@ def pick(context, name, default=None):
     try:
         return context[name]
     except (KeyError, TypeError, AttributeError):
-        if isinstance(name, str) and hasattr(context, name):
-            return getattr(context, name)
+        if isinstance(name, basestring):
+            try:
+                exists = hasattr(context, name)
+            except UnicodeEncodeError:
+                # Python 2 raises UnicodeEncodeError on non-ASCII strings
+                pass
+            else:
+                if exists:
+                    return getattr(context, name)
         if hasattr(context, 'get'):
             return context.get(name)
         return default
@@ -293,6 +313,8 @@ def resolve(context, *segments):
         if segment in (None, ""):
             continue
         if type(context) in (list, tuple):
+            if segment == 'length':
+                return len(context)
             offset = int(segment)
             context = context[offset]
         elif isinstance(context, Scope):
@@ -304,7 +326,22 @@ def resolve(context, *segments):
 
 def resolve_subexpr(helpers, name, context, *args, **kwargs):
     if name not in helpers:
-        raise PybarsError(u"Could not find property %s" % (name,))
+        segments = name.split('..')
+
+        if context is None or len(segments) < 2:
+            raise PybarsError(u"Could not find property %s" % (name,))
+
+        ref = None
+        if type(context) in (list, tuple):
+            ref = context[segments[0]]
+        elif isinstance(context, Scope):
+            ref = context.get(segments[0])
+
+        for segment in segments[1:]:
+            ref = getattr(ref, segment)
+
+        return ref(context, *args, **kwargs)
+
     return helpers[name](context, *args, **kwargs)
 
 
@@ -462,7 +499,7 @@ class FunctionContainer:
             u'import pybars\n'
             u'\n'
             u'if pybars.__version__ != %s:\n'
-            u'    raise pybars.PybarsError("This template was precompiled with pybars3 version %s, running version %%s" %% pybars.__version__)\n'
+            u'    raise pybars.PybarsError("This template was precompiled with pybars4 version %s, running version %%s" %% pybars.__version__)\n'
             u'\n'
             u'from pybars import strlist, Scope, PybarsError\n'
             u'from pybars._compiler import _pybars_, escape, resolve, resolve_subexpr, prepare, ensure_scope\n'
@@ -576,9 +613,9 @@ class CodeBuilder:
                 u"    options['inverse'] = lambda this: None\n"
                 ])
         self._result.grow([
-            u"    value = helper = helpers.get('%s')\n" % symbol,
+            u"    value = helper = helpers.get(u'%s')\n" % symbol,
             u"    if value is None:\n"
-            u"        value = resolve(context, '%s')\n" % symbol,
+            u"        value = resolve(context, u'%s')\n" % symbol,
             u"    if helper and hasattr(helper, '__call__'):\n"
             u"        value = helper(context, options%s\n" % call,
             u"    else:\n"
@@ -608,9 +645,9 @@ class CodeBuilder:
             # XXX: just rm.
             realname = path.replace('.get("', '').replace('")', '')
             self._result.grow([
-                u"    value = helpers.get('%s')\n" % realname,
+                u"    value = helpers.get(u'%s')\n" % realname,
                 u"    if value is None:\n"
-                u"        value = resolve(context, '%s')\n" % path,
+                u"        value = resolve(context, u'%s')\n" % path,
                 ])
         else:
             realname = None
@@ -622,9 +659,14 @@ class CodeBuilder:
         if realname:
             self._result.grow(
                 u"    elif value is None:\n"
-                u"        value = helpers['helperMissing'](context, '%s'%s\n"
+                u"        value = helpers['helperMissing'](context, u'%s'%s\n"
                     % (realname, call)
                 )
+        self._result.grow(
+            u"    if value is None:\n"
+            u"        raise PybarsError(u'Could not find variable %s')\n"
+                % (realname)
+            )
 
     def add_escaped_expand(self, path_type_path, arguments):
         (path_type, path) = path_type_path
@@ -646,8 +688,6 @@ class CodeBuilder:
         self._result.grow(u"    import pdb;pdb.set_trace()\n")
 
     def add_invertedblock(self, symbol, arguments, nested, alt_nested):
-        # This may need to be a blockHelperMissing clal as well.
-
         name = nested.name
         self._locals[name] = nested
 
@@ -673,9 +713,9 @@ class CodeBuilder:
                 u"    options['fn'] = lambda this: None\n"
                 ])
         self._result.grow([
-            u"    value = helper = helpers.get('%s')\n" % symbol,
+            u"    value = helper = helpers.get(u'%s')\n" % symbol,
             u"    if value is None:\n"
-            u"        value = resolve(context, '%s')\n" % symbol,
+            u"        value = resolve(context, u'%s')\n" % symbol,
             u"    if helper and hasattr(helper, '__call__'):\n"
             u"        value = helper(context, options%s\n" % call,
             u"    else:\n"
@@ -683,6 +723,22 @@ class CodeBuilder:
             u"    result.grow(value or '')\n"
             ])
 
+    def add_rawblock(self, symbol, arguments, raw):
+        call = self.arguments_to_call(arguments)
+        self._result.grow([
+            u"    options = {'fn': lambda this: %s}\n" % repr(raw),
+            u"    options['helpers'] = helpers\n"
+            u"    options['partials'] = partials\n"
+            u"    options['root'] = root\n"
+            u"    options['inverse'] = lambda this: None\n"
+            u"    helper = helpers.get(u'%s')\n" % symbol,
+            u"    if helper and hasattr(helper, '__call__'):\n"
+            u"        value = helper(context, options%s\n" % call,
+            u"    else:\n"
+            u"        value = %s\n" % repr(raw),
+            u"    result.grow(value or '')\n"
+            ])
+
     def _invoke_template(self, fn_name, this_name):
         self._result.grow([
             u"    result.grow(",
@@ -699,7 +755,7 @@ class CodeBuilder:
         positional_args = 0
         if arguments:
             for argument in arguments:
-                kwmatch = re.match('(\w+)=(.+)$', argument)
+                kwmatch = re.match(r'(\w+)=(.+)$', argument)
                 if kwmatch:
                     if not overrides:
                         overrides = {}
@@ -755,10 +811,10 @@ class Compiler:
         :return:
             The word
         """
-        boundry = re.search('{{|{|\s|$', source[:position][::-1])
+        boundry = re.search(r'{{|{|\s|$', source[:position][::-1])
         start_offset = boundry.end() if boundry.group(0).startswith('{') else boundry.start()
 
-        boundry = re.search('}}|}|\s|$', source[position:])
+        boundry = re.search(r'}}|}|\s|$', source[position:])
         end_offset = boundry.end() if boundry.group(0).startswith('}') else boundry.start()
 
         return source[position - start_offset:position + end_offset]
@@ -777,9 +833,9 @@ class Compiler:
         if not isinstance(source, str_class):
             raise PybarsError("Template source must be a unicode string")
 
-        tree, (position, _) = self._handlebars(source).apply('template')
+        source = self.whitespace_control(source)
 
-        self.clean_whitespace(tree)
+        tree, (position, _) = self._handlebars(source).apply('template')
 
         if debug:
             print('\nAST')
@@ -803,6 +859,34 @@ class Compiler:
         output = self._compiler(tree).apply('compile')[0]
         return output
 
+    def whitespace_control(self, source):
+        """
+        Preprocess source to handle whitespace control and remove extra block
+        whitespaces.
+
+        :param source:
+            The template source as a unicode string
+        :return:
+            The processed template source as a unicode string
+        """
+        cleanup_sub = re.compile(
+            # Clean-up whitespace control marks and spaces between blocks tags
+            r'(?<={{)~|~(?=}})|(?<=}})[ \t]+(?={{)').sub
+
+        return re.sub(
+            # Whitespace control using "~" mark
+            r'~}}\s*|\s*{{~|'
+
+            # Whitespace around alone blocks tags that in a line
+            r'(?<=\n)([ \t]*{{(#[^{}]+|/[^{}]+|![^{}]+|else|else if [^{}]+)}}[ \t]*)+\r?\n|'
+
+            # Whitespace aroud alone blocks tag on the first line
+            r'^([ \t]*{{(#[^{}]+|![^{}]+)}}[ \t]*)+\r?\n|'
+
+            # Whitespace aroud alone blocks tag on the last line
+            r'\r?\n([ \t]*{{(/[^{}]+|![^{}]+)}}[ \t]*)+$',
+            lambda match: cleanup_sub('', match.group(0).strip()), source)
+
     def precompile(self, source):
         """
         Generates python source code that can be saved to a file for caching
@@ -841,8 +925,7 @@ class Compiler:
             path = path.replace('\\', '/')
             path = path.replace('/', '_')
             mod_name = make_module_name(path)
-            if mod_name in sys.modules:
-                generate_name = True
+            generate_name = mod_name in sys.modules
 
         if generate_name:
             mod_name = make_module_name(path, self.template_counter)
@@ -858,83 +941,14 @@ class Compiler:
 
         return mod.__dict__[container.name]
 
-    def clean_whitespace(self, tree):
-        """
-        Cleans up whitespace around block open and close tags if they are the
-        only thing on the line
-
-        :param tree:
-            The AST - will be modified in place
-        """
-
-        pointer = 0
-        end = len(tree)
-
-        while pointer < end:
-            piece = tree[pointer]
-            if piece[0] == 'block':
-                child_tree = piece[3]
-
-                # Look at open tag, if the only other thing on the line is whitespace
-                # then delete it so we don't introduce extra newlines to the output
-                open_pre_whitespace = False
-                open_pre_content = True
-                if pointer > 1 and tree[pointer - 1][0] == 'whitespace' and (tree[pointer - 2][0] == 'newline' or tree[pointer - 2] == 'template'):
-                    open_pre_whitespace = True
-                    open_pre_content = False
-                elif pointer > 0 and (tree[pointer - 1][0] == 'newline' or tree[pointer - 1] == 'template'):
-                    open_pre_content = False
-
-                open_post_whitespace = False
-                open_post_content = True
-                child_len = len(child_tree)
-                if child_len > 2 and child_tree[1][0] == 'whitespace' and child_tree[2][0] == 'newline':
-                    open_post_whitespace = True
-                    open_post_content = False
-                elif child_len > 1 and child_tree[1][0] == 'newline':
-                    open_post_content = False
-
-                if not open_pre_content and not open_post_content:
-                    if open_pre_whitespace:
-                        tree.pop(pointer - 1)
-                        pointer -= 1
-                        end -= 1
-                    if open_post_whitespace:
-                        child_tree.pop(1)
-                    child_tree.pop(1)  # trailing newline
-
-                # Do the same thing, but for the close tag
-                close_pre_whitespace = False
-                close_pre_content = True
-                child_len = len(child_tree)
-                if child_len > 2 and child_tree[child_len - 1][0] == 'whitespace' and child_tree[child_len - 2][0] == 'newline':
-                    close_pre_whitespace = True
-                    close_pre_content = False
-                elif child_len > 1 and child_tree[child_len - 1][0] == 'newline':
-                    close_pre_content = False
-
-                close_post_whitespace = False
-                close_post_content = True
-                tree_len = len(tree)
-                if tree_len > pointer + 2 and tree[pointer + 1][0] == 'whitespace' and tree[pointer + 2][0] == 'newline':
-                    close_post_whitespace = True
-                    close_post_content = False
-                elif tree_len == pointer + 2 and tree[pointer + 1][0] == 'whitespace':
-                    close_post_whitespace = True
-                    close_post_content = False
-                elif tree_len > pointer + 1 and tree[pointer + 1][0] == 'newline':
-                    close_post_content = False
-                elif tree_len == pointer + 1:
-                    close_post_content = False
-
-                if not close_pre_content and not close_post_content:
-                    if close_pre_whitespace:
-                        child_tree.pop()
-                    child_tree.pop()  # preceeding newline
-                    if close_post_whitespace:
-                        tree.pop(pointer + 1)
-                        end -= 1
-
-                self.clean_whitespace(child_tree)
-
-            pointer += 1
+    def template(self, code):
+        def _render(context, helpers=None, partials=None, root=None):
+            ns = {
+                'context': context,
+                'helpers': helpers,
+                'partials': partials,
+                'root': root
+            }
+            exec(code + '\nresult = render(context, helpers=helpers, partials=partials, root=root)', ns)
+            return ns['result']
+        return _render
